<?php
namespace Grav\Plugin;

use \Grav\Common\Grav;

class TwigPcreExtension extends \Twig_Extension
{

    protected $config;

    public function __construct()
    {
        $this->config = Grav::instance()['config'];
    }

    /**
     * Returns extension name.
     *
     * @return string
     */
    public function getName()
    {
        return 'TwigPcreExtension';
    }

    /**
     * Register twig filter.
     *
     * @return array
     */
    public function getFilters()
    {
        return array(
            'preg_quote' => new \Twig_SimpleFilter('preg_quote', array($this, 'quote')),
            'preg_match' => new \Twig_SimpleFilter('preg_match', array($this, 'match')),
            'preg_get' => new \Twig_SimpleFilter('preg_get', array($this, 'get')),
            'preg_get_all' => new \Twig_SimpleFilter('preg_get_all', array($this, 'getAll')),
            'preg_grep' => new \Twig_SimpleFilter('preg_grep', array($this, 'grep')),
            'preg_replace' => new \Twig_SimpleFilter('preg_replace', array($this, 'replace')),
            'preg_filter' => new \Twig_SimpleFilter('preg_filter', array($this, 'filter')),
            'preg_split' => new \Twig_SimpleFilter('preg_splig', array($this, 'split')),
        );
    }

    /**
     * Check that the regex doesn't use the eval modifier
     * 
     * @param string $pattern
     */
    protected function assertNoEval($pattern)
    {
        if (preg_match('/(.).*\1(.+)$/', trim($pattern), $match) && strpos($match[1], 'e') !== false) throw new \Exception("Using the eval modifier for regular expressions is not allowed");
    }
    

    /**
     * Quote regular expression characters.
     * 
     * @param string $value
     * @param string $delimiter
     * @return string
     */
    public function quote($value, $delimiter = '/')
    {
        if (!isset($value)) return null;
        return preg_quote($value, $delimiter);
    }

    /**
     * Perform a regular expression match.
     * 
     * @param string $value
     * @param string $pattern
     * @return boolean
     */
    public function match($value, $pattern)
    {
        $this->assertNoEval($pattern);
        
        if (!isset($value)) return null;
        return preg_match($pattern, $value);
    }

    /**
     * Perform a regular expression match and return a matched group.
     * 
     * @param string $value
     * @param string $pattern
     * @return string
     */
    public function get($value, $pattern, $group=0)
    {
        $this->assertNoEval($pattern);
        
        if (!isset($value)) return null;
        if (!preg_match($pattern, $value, $matches)) return null;
        return isset($matches[$group]) ? $matches[$group] : null;
    }

    /**
     * Perform a regular expression match and return the group for all matches.
     * 
     * @param string $value
     * @param string $pattern
     * @return array
     */
    public function getAll($value, $pattern, $group=0)
    {
        $this->assertNoEval($pattern);
        
        if (!isset($value)) return null;
        if (!preg_match_all($pattern, $value, $matches, PREG_PATTERN_ORDER)) return array();
        return isset($matches[$group]) ? $matches[$group] : array();
    }

    /**
     * Perform a regular expression match and return an array of entries that match the pattern
     * 
     * @param array  $values
     * @param string $pattern
     * @param strign $flags    Optional 'invert' to return entries that do not match the given pattern.
     * @return array
     */
    public function grep($values, $pattern, $flags='')
    {
        $this->assertNoEval($pattern);
        
        if (!isset($values)) return null;
        
        if (is_string($flags)) $flags = $flags == 'invert' ? PREG_GREP_INVERT : 0;
        return preg_grep($pattern, $values, $flags);
    }

    /**
     * Perform a regular expression search and replace.
     * 
     * @param string $value
     * @param string $pattern
     * @param string $replacement
     * @param int    $limit
     * @return string
     */
    public function replace($value, $pattern, $replacement='', $limit=-1)
    {
        $this->assertNoEval($pattern);
        
        if (!isset($value)) return null;
        return preg_replace($pattern, $replacement, $value, $limit);
    }

    /**
     * Perform a regular expression search and replace, returning only matched subjects.
     * 
     * @param string $value
     * @param string $pattern
     * @param string $replacement
     * @param int    $limit
     * @return string
     */
    public function filter($value, $pattern, $replacement='', $limit=-1)
    {
        $this->assertNoEval($pattern);
        
        if (!isset($value)) return null;
        return preg_filter($pattern, $replacement, $value, $limit);
    }

    /**
     * Split text into an array using a regular expression.
     * 
     * @param string $value
     * @param string $pattern
     * @return array
     */
    public function split($value, $pattern)
    {
        $this->assertNoEval($pattern);
        
        if (!isset($value)) return null;
        return preg_split($pattern, $value);
    }

}
